import matplotlib.pyplot as pl
import matplotlib.image as image
import matplotlib.ticker as mticker

import numpy as np
import os
import json
from pathlib import Path

import numpy as np
import emcee
import corner
from matplotlib import rc

rc("font", **{"family": "sans-serif", "sans-serif": ["Helvetica"]})
rc("text")

from .ConfigReader import ConfigReader
from .PredictionBuilder import PredictionBuilder


class SummaryPlotter:
    def __init__(
        self,
        config: ConfigReader,
        pb: PredictionBuilder,
        sampler: emcee.EnsembleSampler,
    ):

        samples = sampler.chain.reshape(-1, len(config.prior_limits))

        # make directory to hold results of this run
        run_name = config.params["config"]["run_name"]
        results_path = Path("results") / run_name
        results_path.mkdir(parents=True, exist_ok=True)

        mcmc_params = np.mean(sampler.flatchain, axis=0)
        mcmc_params_cov = np.cov(np.transpose(sampler.flatchain))

        data_label = f"Data ({run_name})"
        max_val = (1.5) * (max(config.params["config"]["data"]["central_values"]))
        min_val = (0.0) * (min(config.params["config"]["data"]["central_values"]))
        xlabel = config.observable
        ylabel = "d$\sigma_{tt}$/dX"

        # make plots of prediction at max point of posterior versus data
        coefficients = list(config.coefficients)
        pl.figure()
        label_string_bestfit = "best-fit: ("
        for c in range(0, len(config.coefficients)):
            if c == (len(config.coefficients) - 1):
                label_string_bestfit = (
                    label_string_bestfit
                    + coefficients[c]
                    + " = "
                    + "%.1f" % mcmc_params[c]
                    + ")"
                )
            else:
                label_string_bestfit = (
                    label_string_bestfit
                    + coefficients[c]
                    + " = "
                    + "%.1f" % mcmc_params[c]
                    + ", "
                )

        pl.errorbar(
            config.x_vals,
            pb.make_prediction(mcmc_params),
            fmt="m",
            xerr=0.0,
            yerr=0.0,
            label=label_string_bestfit,
        )
        pl.errorbar(
            config.x_vals,
            config.params["config"]["data"]["central_values"],
            fmt="o",
            xerr=0.25,
            yerr=0.05,
            label=data_label,
        )
        pl.axis(
            [
                config.x_vals[0] - 0.25,
                config.x_vals[len(config.x_vals) - 1] + 0.25,
                min_val,
                max_val,
            ]
        )
        ax = pl.gca()
        labely = ax.set_xlabel(xlabel, fontsize=18)
        labely = ax.set_ylabel(ylabel, fontsize=18)
        ax.xaxis.set_label_coords(0.85, -0.065)
        ax.yaxis.set_label_coords(-0.037, 0.83)
        pl.legend(loc=2)
        pl.savefig(results_path / f"{run_name}_bestfit_predictions.png")
        pl.close()

        #  make "corner" plot
        labels = []
        ranges = []
        for c in config.params["config"]["model"]["prior_limits"].keys():
            label = "$" + c + "$"
            labels.append(label)
            ranges.append(config.params["config"]["model"]["prior_limits"][c])

        fig = corner.corner(
            samples,
            labels=config.tex_labels,
            label_kwargs={"fontsize": 18},
            range=ranges,
            quantiles=[0.16, 0.84],
            levels=(1 - np.exp(-0.5),),
            truths=np.zeros(len(labels)),
            show_titles=True,
            title_kwargs={"fontsize": 18},
        )

        plotfilename = results_path / f"{run_name}.png"

        ax0 = fig.add_subplot(999)
        ax0.axis("off")

        fig.savefig(plotfilename)
        pl.close()

        bestFits = []
        marginUncsUp = []
        marginUncsDown = []
        x = []

        for i in range(len(labels)):
            mcmc = np.percentile(samples[:, i], [16, 50, 84])
            q = np.diff(mcmc)
            txt = "\mathrm{{{3}}} = {0:.3f}_{{-{1:.3f}}}^{{{2:.3f}}}"
            txt = txt.format(mcmc[1], q[0], q[1], labels[i])
            x.append(i)
            bestFits.append(mcmc[1])
            marginUncsUp.append(q[0])
            marginUncsDown.append(q[1])

        pl.figure()
        pl.tight_layout()
        pl.gcf().subplots_adjust(bottom=0.15)

        fig = pl.errorbar(
            x,
            bestFits,
            yerr=[marginUncsDown, marginUncsUp],
            fmt="o",
            label=r"posterior median and CI",
        )
        pl.xticks(range(len(labels)), config.tex_labels, rotation="45", fontsize=21)
        ax = pl.gca()
        ax.yaxis.set_label_coords(-0.075, 0.83)
        labely = ax.set_ylabel(r"$c_{i}$ [$GeV^{-2}$]", fontsize=18)
        pl.legend(loc=1, fontsize=14)
        pl.savefig(results_path / f"{run_name}_bestFits.png")
        pl.close()

        ############################################
        ######## Corner with 1-d CL overlaid  ######
        ############################################

        fig_overlay = corner.corner(
            samples,
            labels=config.tex_labels,
            label_kwargs={"fontsize": 21},
            range=ranges,
            color="k",
            quantiles=[0.16, 0.84],
            levels=(1 - np.exp(-0.5),),
            truths=np.zeros(len(labels)),
            show_titles=True,
            title_kwargs={"fontsize": 19},
            hist2d_kwargs={"fill_contours": True, "plot_density": True},
        )

        resplot = image.imread(results_path / f"{run_name}_bestFits.png")

        ax0 = fig_overlay.add_subplot(322)
        ax0.axis("off")
        img = ax0.imshow(resplot)

        ax0 = fig_overlay.add_subplot(5, 5, 15)
        ax0.axis("off")

        plotfilename = results_path / f"{run_name}_overlay.png"

        fig_overlay.savefig(plotfilename)
        pl.close()

        # make plots of best fit prediction versus data
        pl.figure()
        label_string_bestfit = "best-fit prediction"

        pl.errorbar(
            config.x_vals,
            pb.make_prediction(bestFits),
            fmt="m",
            xerr=0.0,
            yerr=0.0,
            label=label_string_bestfit,
        )
        pl.errorbar(
            config.x_vals,
            config.params["config"]["data"]["central_values"],
            fmt="o",
            xerr=0.25,
            yerr=np.sqrt(
                np.diagonal(config.params["config"]["data"]["covariance_matrix"])
            ),
            label=data_label,
        )
        pl.axis(
            [
                config.x_vals[0] - 0.25,
                config.x_vals[len(config.x_vals) - 1] + 0.25,
                min_val,
                max_val,
            ]
        )
        ax = pl.gca()
        labely = ax.set_xlabel(xlabel, fontsize=18)
        labely = ax.set_ylabel(ylabel, fontsize=18)
        ax.xaxis.set_label_coords(0.85, -0.065)
        ax.yaxis.set_label_coords(-0.037, 0.83)
        pl.legend(loc=1)
        pl.savefig(results_path / f"{run_name}_bestfit_predictions.png")
        pl.close()

        # make fit summary json
        fitSummary = {}
        fitSummary["bestFit"] = bestFits
        fitSummary["bestFit"] = bestFits
        fitSummary["UncsUp"] = marginUncsUp
        fitSummary["UncsDown"] = marginUncsDown
        fitSummary["labels"] = labels
        fitSummary["x"] = x

        with open(run_name + ".json", "w") as fs:
            json.dump(fitSummary, fs)

        pl.figure()

        # best-fit prediction and random samples compared to data
        uncX = np.absolute((np.diff(config.bins))) / 2.0
        dataUncY = np.sqrt(np.diag(config.cov))
        pl.errorbar(
            config.x_vals,
            config.params["config"]["data"]["central_values"],
            fmt="o",
            xerr=uncX,
            yerr=dataUncY,
            label=data_label,
        )

        inds = np.random.randint(len(samples), size=499)
        for ind in inds:
            sample = samples[ind]
            pl.plot(config.x_vals, pb.make_prediction(sample), "C1", alpha=0.02)

        pl.plot(
            config.x_vals,
            pb.make_prediction(samples[42]),
            "C1",
            label="500 random samples",
            alpha=0.1,
        )

        ax = pl.gca()
        labelx = ax.set_xlabel(xlabel, fontsize=18)
        labely = ax.set_ylabel(ylabel, fontsize=18)
        ax.xaxis.set_label_coords(0.85, -0.065)
        ax.yaxis.set_label_coords(-0.037, 0.83)
        pl.legend(loc=1)
        pl.savefig(results_path / f"{run_name}_postfit_predictions.png")
        pl.close()

        pl.figure()
        fig, axes = pl.subplots(len(labels), figsize=(10, 7), sharex=True)
        samples = sampler.get_chain()
        for i in range(len(labels)):
            ax = axes[i]
            ax.plot(samples[:, :, i], "k", alpha=0.3)
            ax.set_xlim(0, len(samples))
            ax.set_ylabel(labels[i])
            ax.yaxis.set_label_coords(-0.1, 0.5)

        axes[-1].set_xlabel("step number")

        pl.savefig(results_path / f"{run_name}_walkersPaths.png")
        pl.close()

        # covariance matrix of coefficicents
        fig, ax = pl.subplots()
        im = ax.imshow(mcmc_params_cov, cmap=pl.cm.Blues)

        ax.set_xticks(np.arange(len(config.tex_labels)))
        ax.set_yticks(np.arange(len(config.tex_labels)))
        # ... and label them with the respective list entries
        ax.set_xticklabels(config.tex_labels)
        ax.set_yticklabels(config.tex_labels)

        # fixing yticks with matplotlib.ticker "FixedLocator"
        # ticks_loc = ax.get_yticks().tolist()
        # ticks_loc = config.tex_labels

        # ax.yaxis.set_major_locator(mticker.FixedLocator(ticks_loc))
        # label_format = '{:,.0f}'

        # pl.imshow(mcmc_params_cov, cmap=pl.cm.Blues)
        # ax.set_xticks(np.arange(len(config.tex_labels)))

        # ax.set_yticklabels(ticks_loc)

        pl.savefig(results_path / f"{run_name}_mcmc_params_cov.png")
        pl.close()

        # Write SM pred to text file for validation
        sm_file_name = results_path / f"{run_name}_postFit_sm_pred.txt"
        f = open(sm_file_name, "w")
        sm_pred = str(
            repr(
                pb.make_prediction(
                    np.zeros(len(config.params["config"]["model"]["prior_limits"]))
                )
            )
        )
        f.write(sm_pred)
        f.close()



import numpy as np
import itertools
from sklearn.linear_model import LinearRegression

from typing import List


def triangular_number(n: int) -> int:
    return int(n * (n + 1) / 2)


class PredictionBuilder:
    """
    Constructs Linear Regression Model for the wilson coefficient of each operator for some observable
    """

    def __init__(self, nOps: int, samples: List[float], preds: List[List[float]]):
        """ Constructor for PredictionBuilder """

        self.nOps = nOps
        self.nSamples = int((nOps + 1) * (nOps + 2) / 2)
        self.model = self.build_regression_model(nOps, samples, preds)

    def build_regression_model(
        self, nOps: int, samples: List[float], preds: List[List[float]]
    ) -> LinearRegression:
        """ Initialise morphing model using samples with predicted values """
        if len(preds) < self.nSamples:
            raise TypeError(
                "morphing with "
                + str(nOps)
                + " coefficients requires at least "
                + str(self.nSamples)
                + " samples,  but only "
                + str(len(preds))
                + " are provided"
            )

        # convert to vector of coefficient factors to linearise the morphing
        cInputAr = self.make_coefficients(samples)

        # define model
        model = LinearRegression()

        # fit model
        model.fit(cInputAr, preds)

        return model

    def make_coefficients(self, ci: List[float]) -> np.ndarray:
        X = np.array([])
        num_rows = np.shape(ci)[0]
        for row in ci:
            # Account for quadratic self term
            X = np.append(X, row ** 2)

            # Account for quadratic cross term
            combs = itertools.combinations(list(row), 2)
            for comb in combs:
                X = np.append(X, comb[0] * comb[1])
        X = X.reshape(num_rows, triangular_number(len(ci[0])))
        return X

    def makeCoeffPoint(self, ci: np.ndarray) -> np.ndarray:
        X = np.array([])
        X = np.append(X, ci ** 2)
        combs = itertools.combinations(list(ci), 2)
        for comb in combs:
            X = np.append(X, comb[0] * comb[1])

        X = X.reshape(1, triangular_number(len(ci)))
        return X

    def make_prediction(self, c: np.ndarray) -> np.ndarray:
        """ Produce the predicted observable for a set of coefficients (excluding SM coefficient) """
        c = np.append(1.0, c)
        cInputAr = self.makeCoeffPoint(c)
        pred = self.model.predict(cInputAr)
        return pred[0]

    # TODO: Not in use. Figure out if it can be deleted
    # def init(self, nOps, samples, preds):
    #     # 'exact' morphing, soon to be defunct

    #     if len(preds) <= self.nSamples:
    #         raise TypeError(
    #             "morphing with "
    #             + str(nOps)
    #             + " coefficients requires at least "
    #             + str(self.nSamples)
    #             + " samples,  but only "
    #             + str(len(preds))
    #             + " are provided"
    #         )

    #     self.nOps = nOps
    #     cInputAr = np.array([])

    #     for sample in samples:
    #         couplings = np.array([])
    #         combs = list(itertools.combinations(sample, 2))
    #         print("sample = " + str(sample))

    #         # SM term
    #         couplings = np.append(couplings, sample[0] ** 2)
    #         #'SM--Oi' interference terms
    #         for comb in range(0, nOps):
    #             couplings = np.append(couplings, combs[comb][0] * combs[comb][1])
    #         #'squared' terms
    #         for ci in range(1, nOps + 1):
    #             couplings = np.append(couplings, sample[ci] ** 2)
    #         #'cross' terms
    #         for comb in range(nOps, len(combs)):
    #             couplings = np.append(couplings, combs[comb][0] * combs[comb][1])

    #         cInputAr = np.append(cInputAr, couplings, axis=0)

    #     cInputAr = np.reshape(cInputAr, (len(samples), len(samples)))
    #     cInputAr = np.transpose(cInputAr)
    #     cInputAr.tofile("cInputAr.txt", sep=" ", format="%s")

    #     inWeightsAr = np.linalg.inv(cInputAr)

    #     self.inWeightsAr = inWeightsAr
    #     inPreds = preds.transpose()
    #     self.inPreds = inPreds
    #     inPreds.tofile("inPreds.txt", sep=" ", format="%s")



import sys
import json
import numpy as np
from numpy.linalg import inv


class ConfigReader:
    """ Reads JSON configuration file"""

    def __init__(self, filename: str):
        self.filename = filename
        coefficients = []
        predictions = {}
        with open(filename, "r") as f:
            config = json.load(f)

        try:
            self.params = config
            self.run_name = self.params["config"]["run_name"]
            self.observable = self.params["config"]["data"]["observable"]
            self.bins = self.params["config"]["data"]["bins"]
            self.data = self.params["config"]["data"]["central_values"]
            self.prior_limits = self.params["config"]["model"]["prior_limits"]
            self.coefficients = list(
                self.params["config"]["model"]["prior_limits"].keys()
            )
            self.n_walkers = self.params["config"]["fit"]["n_walkers"]
            self.n_burnin = self.params["config"]["fit"]["n_burnin"]
            self.n_total = self.params["config"]["fit"]["n_total"]
            self.cov = self.params["config"]["data"]["covariance_matrix"]
            self.icov = inv(self.cov)
            self.x_vals = self.params["config"]["data"]["bins"]
            self.samples = np.asarray(self.params["config"]["model"]["samples"])
            if self.params["config"]["model"]["input"] == "numpy":
                self.predictions = np.asarray(
                    self.params["config"]["model"]["predictions"]
                )
                if len(self.predictions) != len(self.samples):
                    raise TypeError(
                        "number of samples must equal number of predictions"
                    )
        except KeyError as err:
            print(
                f"Error reading file {filename}: Could not find option {err}",
                file=sys.stderr,
            )
            exit(0)

        self.tex_labels = ["$" + c + "$" for c in self.coefficients]
        x_vals = np.zeros(len(self.x_vals) - 1)
        for x_val in range(0, len(self.bins) - 1):
            x_vals[x_val] = (self.bins[x_val] + self.bins[x_val + 1]) / 2.0
        self.x_vals = x_vals



from multiprocessing import Pool

import numpy as np
import emcee

from .ConfigReader import ConfigReader
from .PredictionBuilder import PredictionBuilder

_PB = None
_DATA = None
_ICOV = None
_CONFIG = None

def ln_prior(c: np.ndarray, config: ConfigReader) -> float:
    lnp = 0.0
    for scan_ind in range(0, len(c)):
        if (c[scan_ind] < config.prior_limits[config.coefficients[scan_ind]][0]) | (
            c[scan_ind] > config.prior_limits[config.coefficients[scan_ind]][1]
        ):
            lnp = -np.inf
    return lnp


def ln_prob(
    c: np.ndarray,
) -> float:
    pred = _PB.make_prediction(c)
    diff = pred - _DATA
    ll = (-np.dot(diff, np.dot(_ICOV, diff))) + (ln_prior(c, _CONFIG))
    return ll


class MCMCFitter:
    """ Use Markov chain Monte Carlo method to fit the model from the PredictionBuilder to some data specified in the configuration file """

    def __init__(
        self,
        config: ConfigReader,
        pb: PredictionBuilder,
        use_multiprocessing: bool = True,
    ):
        n_walkers = config.n_walkers

        n_dim = int(len(config.prior_limits))
        n_burnin = config.n_burnin
        n_total = config.n_total
        p0 = [np.zeros(n_dim) + 1e-4 * np.random.randn(n_dim) for i in range(n_walkers)]

        global _DATA
        global _ICOV
        global _CONFIG
        global _PB
        _DATA = config.params["config"]["data"]["central_values"]
        _ICOV = config.icov
        _CONFIG = config
        _PB = pb

        if use_multiprocessing:
            with Pool() as pool:
                sampler = emcee.EnsembleSampler(
                    n_walkers,
                    n_dim,
                    ln_prob,
                    pool=pool,
                )
                # Run burn in runs
                pos, prob, state = sampler.run_mcmc(p0, n_burnin, progress=True)
                sampler.reset()

                # Perform proper run
                sampler.run_mcmc(pos, n_total, progress=True)
        else:
            sampler = emcee.EnsembleSampler(
                n_walkers,
                n_dim,
                ln_prob,
            )
            # Run burn in runs
            pos, prob, state = sampler.run_mcmc(p0, n_burnin, progress=True)
            sampler.reset()

            # Perform proper run
            sampler.run_mcmc(pos, n_total, progress=True)

        self.sampler = sampler
